#!/usr/bin/env python
"""XXX
"""

from __future__ import print_function
from ZODB import DB
from ZODB.MappingStorage import MappingStorage
from ZODB.POSException import ConflictError
import transaction
from persistent import Persistent
from random import randint

from golang import func, defer, select, default
from golang import sync, context


# PInt is persistent integer.
class PInt(Persistent):
    def __init__(self, i):
        self.i = i

from wendelin.lib.testing import TestDB_ZEO


@func
def main():
    """
    if 0:
        zstor = MappingStorage()
        db = DB(zstor)
    else:
        tdb = TestDB_ZEO('<zeo>')
        tdb.setup()
        defer(tdb.teardown)

        zstor = tdb.getZODBStorage()
        db = DB(zstor)
        #zstor.app.poll_thread.name = 'C.poll'
    """

    tdb = TestDB_ZEO('<zeo>')
    tdb.setup()
    defer(tdb.teardown)

    def dbopen():
        zstor = tdb.getZODBStorage()
        db = DB(zstor)
        return db


    # init initializes the database with two integer objects - obj1/obj2 that are set to 0.
    @func
    def init():
        db = dbopen()
        defer(db.close)

        transaction.begin()
        zconn = db.open()

        root = zconn.root()
        root['obj1'] = PInt(0)
        root['obj2'] = PInt(0)

        transaction.commit()
        zconn.close()


    # T1 accesses obj1/obj2 in a loop and verifies that obj1.i == obj2.i
    #
    # access to obj1 is organized to always trigger loading from zstor.
    # access to obj2 goes through zconn cache and so verifies whether the cache is not stale.
    #
    # XXX + commit
    @func
    def T1(ctx, name, N):
        db = dbopen()
        defer(db.close)

        def t1():
            transaction.begin()
            zconn = db.open()

            root = zconn.root()
            obj1 = root['obj1']
            obj2 = root['obj2']

            # obj1 - reload it from zstor
            # obj2 - get it from zconn cache
            obj1._p_invalidate()

            # both objects must have the same values
            i1 = obj1.i
            i2 = obj2.i
            if i1 != i2:
                print('FAIL')
                raise AssertionError("T%s: obj1.i (%d)  !=  obj2.i (%d)" % (name, i1, i2))

            # change objects once in a while
            if randint(0,4) == 0:
                #print("T%s: modify" % name)
                obj1.i += 1
                obj2.i += 1

            #transaction.abort() # we did not changed anything; also fails with commit
            try:
                transaction.commit()
            except ConflictError:
                #print('conflict -> ignore')
                transaction.abort()

            zconn.close()

        for i in range(N):
            if ready(ctx.done()):
                break
            #print('T%s.%d' % (name, i))
            t1()

        raise RuntimeError("T1: done")


    """
    # T2 changes obj1/obj2 in a loop by doing `objX.i += 1`.
    #
    # Since both objects start from 0, the invariant that `obj1.i == obj2.i` is always preserved.
    @func
    def T2(ctx, N):
        db = dbopen()
        defer(db.close)

        def t2():
            transaction.begin()
            zconn = db.open()

            root = zconn.root()
            obj1 = root['obj1']
            obj2 = root['obj2']
            obj1.i += 1
            obj2.i += 1
            assert obj1.i == obj2.i

            transaction.commit()
            zconn.close()

        for i in range(N):
            if ready(ctx.done()):
                break
            #print('T2.%d' % i)
            t2()
    """


    # run T1 and T2 concurrently. As of 20191210, due to race condition in
    # Connection.open, it triggers the bug where T1 sees stale obj2 with obj1.i != obj2.i
    init()

    N = 100000
    wg = sync.WorkGroup(context.background())
    for x in 'abcdefgh':
        wg.go(T1, x, N)
    #wg.go(T2, N)
    wg.wait()

    print('OK')


# ready returns whether channel ch is ready.
def ready(ch):
    _, _rx = select(
        default,        # 0
        ch.recv,        # 1
    )
    if _ == 0:
        return False
    return True

if __name__ == '__main__':
    main()
